@@ -2,19 +2,7 @@
  * @file encode.c
  *
  * @section LICENSE
- * Copyright 2012 - 2019
- *
- * Mathis Rosenhauer, Moritz Hanke, Joerg Behrens
- * Deutsches Klimarechenzentrum GmbH
- * Bundesstr. 45a
- * 20146 Hamburg Germany
- *
- * Luis Kornblueh
- * Max-Planck-Institut fuer Meteorologie
- * Bundesstr. 53
- * 20146 Hamburg
- * Germany
- *
+ * Copyright 2023 Mathis Rosenhauer, Moritz Hanke, Joerg Behrens, Luis Kornblueh
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,7 +32,7 @@
  * @section DESCRIPTION
  *
  * Adaptive Entropy Encoder
- * Based on CCSDS documents 121.0-B-2 and 120.0-G-3
+ * Based on the CCSDS recommended standard 121.0-B-3
  *
  */
 
@@ -327,7 +315,7 @@ static uint32_t assess_splitting_option(struct aec_stream *strm)
        In Rice coding each sample in a block of samples is split at
        the same position into k LSB and bits_per_sample - k MSB. The
        LSB part is left binary and the MSB part is coded as a
-       fundamental sequence a.k.a. unary (see CCSDS 121.0-B-2). The
+       fundamental sequence a.k.a. unary (see CCSDS 121.0-B-3). The
        function of the length of the Coded Data Set (CDS) depending on
        k has exactly one minimum (see A. Kiely, IPN Progress Report
        42-159).
@@ -339,7 +327,7 @@ static uint32_t assess_splitting_option(struct aec_stream *strm)
        increases, then the minimum will be found with smaller k.
 
        For increasing k we know that we will gain block_size bits in
-       length through the larger binary part. If the FS lenth is less
+       length through the larger binary part. If the FS length is less
        than the block size then a reduced FS part can't compensate the
        larger binary part. So we know that the CDS for k+1 will be
        larger than for k without actually computing the length. An
@@ -423,6 +411,7 @@ static uint32_t assess_se_option(struct aec_stream *strm)
         if (len > state->uncomp_len)
             return UINT32_MAX;
     }
+
     return (uint32_t)len;
 }
 
@@ -489,11 +478,10 @@ static int m_flush_block(struct aec_stream *strm)
     struct internal_state *state = strm->state;
 
 #ifdef ENABLE_RSI_PADDING
-    if (state->blocks_avail == 0
-        && strm->flags & AEC_PAD_RSI
-        && state->block_nonzero == 0
-        )
-        emit(state, 0, state->bits % 8);
+    if (state->blocks_avail == 0 &&
+        strm->flags & AEC_PAD_RSI &&
+        state->block_nonzero == 0)
+            emit(state, 0, state->bits % 8);
 #endif
 
     if (state->direct_out) {
@@ -501,6 +489,14 @@ static int m_flush_block(struct aec_stream *strm)
         strm->next_out += n;
         strm->avail_out -= n;
         state->mode = m_get_block;
+
+        if (state->ready_to_capture_rsi &&
+            state->blocks_avail == 0 &&
+            state->offsets != NULL) {
+                vector_push_back(state->offsets, (strm->total_out - strm->avail_out) * 8 + (8 - state->bits));
+                state->ready_to_capture_rsi = 0;
+        }
+
         return M_CONTINUE;
     }
 
@@ -653,7 +649,7 @@ static int m_get_rsi_resumable(struct aec_stream *strm)
     /**
        Get RSI while input buffer is short.
 
-       Let user provide more input. Once we got all input pad buffer
+       Let user provide more input. Once we got all input, pad buffer
        to full RSI.
     */
 
@@ -668,6 +664,8 @@ static int m_get_rsi_resumable(struct aec_stream *strm)
                     state->blocks_avail = state->i / strm->block_size - 1;
                     if (state->i % strm->block_size)
                         state->blocks_avail++;
+                    /* Pad raw buffer with last sample. Only encode
+                     * blocks_avail will be encoded later. */
                     do
                         state->data_raw[state->i] =
                             state->data_raw[state->i - 1];
@@ -719,8 +717,8 @@ static int m_get_block(struct aec_stream *strm)
         state->blocks_avail = strm->rsi - 1;
         state->block = state->data_pp;
         state->blocks_dispensed = 1;
-
         if (strm->avail_in >= state->rsi_len) {
+            state->ready_to_capture_rsi = 1;
             state->get_rsi(strm);
             if (strm->flags & AEC_DATA_PREPROCESS)
                 state->preprocess(strm);
@@ -892,7 +890,8 @@ int aec_encode_init(struct aec_stream *strm)
     *state->cds = 0;
     state->bits = 8;
     state->mode = m_get_block;
-
+    struct vector_t *offsets = NULL;
+    state->ready_to_capture_rsi = 0;
     return AEC_OK;
 }
 
@@ -931,10 +930,51 @@ int aec_encode_end(struct aec_stream *strm)
     int status = AEC_OK;
     if (state->flush == AEC_FLUSH && state->flushed == 0)
         status = AEC_STREAM_ERROR;
+    if (state->offsets != NULL) {
+        vector_destroy(state->offsets);
+        state->offsets = NULL;
+    }
     cleanup(strm);
     return status;
 }
 
+int aec_encode_count_offsets(struct aec_stream *strm, size_t *count)
+{
+    struct internal_state *state = strm->state;
+    if (state->offsets == NULL) {
+        *count = 0;
+        return AEC_RSI_OFFSETS_ERROR;
+    } else {
+        *count = vector_size(state->offsets);
+    }
+    return AEC_OK;
+}
+
+int aec_encode_get_offsets(struct aec_stream *strm, size_t *offsets, size_t offsets_count)
+{
+    struct internal_state *state = strm->state;
+    if (state->offsets == NULL) {
+        return AEC_RSI_OFFSETS_ERROR;
+    }
+    if (offsets_count < vector_size(state->offsets)) {
+        return AEC_MEM_ERROR;
+    }
+    memcpy(offsets, vector_data(state->offsets), offsets_count * sizeof(size_t));
+    return AEC_OK;
+}
+
+int aec_encode_enable_offsets(struct aec_stream *strm)
+{
+    struct internal_state *state = strm->state;
+
+    if (state->offsets != NULL)
+        return AEC_RSI_OFFSETS_ERROR;
+
+    state->offsets = vector_create();
+    vector_push_back(state->offsets, 0);
+    return AEC_OK;
+}
+
 int aec_buffer_encode(struct aec_stream *strm)
 {
     int status = aec_encode_init(strm);
